/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#include "common.h"
#include "pio_buffer.h"

#include <stdarg.h>
#include <stdio.h>

void buf_create (pio_buffer *buf, size_t size)
{
  if (size == 0)
    buf->buf_fer = NULL;
  else {
    size = __max (1, size);
    if ((buf->buf_fer = calloc (sizeof (char), size)) == NULL) {
      failwith ("buf_create: out of memory");
    }
  }
  buf->buf_length = size;
  buf->buf_position = 0;
}

/* sub, blit, length, reset, add_substring, add_buffer: useless */

/* nth, clear, add_substitute, add_chan, output_buffer: not impl */

size_t ensure (pio_buffer *buf, size_t more)
{
  more = __max (1, more);		/* in case avail + incr == 0 in buf_printf */

  size_t next_position = buf->buf_position + more;
  if (next_position > buf->buf_length) {
    /* resize */
    size_t new_len = buf->buf_length;

    while (next_position > new_len)
      /* XXX new_len *= 2 */
      new_len = next_position;		/* for debugging */
    buf->buf_fer = realloc (buf->buf_fer, new_len);
    if (buf->buf_fer == NULL)
      failwith ("buf_add*/printf: out of memory"); /* XXX voir les ensure */
    buf->buf_length = new_len;
  }
  
  return next_position;
}

void buf_add_char (pio_buffer *buf, char c)
{
  size_t next_position = ensure (buf, sizeof (c));

  buf->buf_fer[buf->buf_position] = c;
  buf->buf_position = next_position;
}

void buf_add_chars_w_len (pio_buffer *buf, char *s, size_t len)
{
  size_t next_position = ensure (buf, len);

  memmove (&buf->buf_fer[buf->buf_position], s, len);
  buf->buf_position = next_position;
}

/* became a #define
void buf_add_chars (pio_buffer *buf, char *s)
{
  buf_add_chars_w_len (buf, s, strlen (s));
  } */

void buf_printf (pio_buffer *buf, char *format, ...)
{
  va_list args;
  const size_t pos = buf->buf_position;

  while (1) {
    va_start (args, format);
    const size_t avail = buf->buf_length - pos; /* may change after ensure below */
    const size_t incr = vsnprintf (&buf->buf_fer[pos], avail, format, args);

    if (incr < avail) {
      buf->buf_position = pos + incr;
      break;
    }

    (void) ensure (buf, avail + incr);	/* should be > 0 */
    va_end (args);			/* we have to restart vsnprintf */
  }
}

struct string buf_contents (pio_buffer *buf) {
  struct string str;
  size_t sz = buf->buf_position;
  char *s = malloc (sz + 1);		/* 1 for final '\0' */
  if (s == NULL)
    failwith (__FILE__"::buf_contents: out of memory");
  memcpy (s, buf->buf_fer, sz);
  s[sz] = '\0';				/* C convenience */
  str.ssize = sz;
  str.str = s;
  return str;
}

void write_delimited_string (pio_buffer *ob, struct string s) {
  buf_add_char (ob, '"');

  unsigned char c;
  size_t i = 0;
  const char *sp = s.str;

  while ((c = (unsigned char)*sp++, ++i <= s.ssize)) {
    if (c > 31 && c < 127)
      buf_add_char (ob, c);
    else {
      buf_add_char (ob, '\\');
      if (c == '\\' || c == '"')
	buf_add_char (ob, c);
      else {
	buf_printf (ob, "%03hhu", c);
      }
    }
  }

  buf_add_char (ob, '"');
}
